! ---------------------------------------------------------------------------
!           Lt_L_ng_3D.f90 - by Alejandro Valenciano 
!                        
!           January 2006.
! ---------------------------------------------------------------------------
!  Parameters: 
!              nax     x offset from the diagonal  (odd)
!              nay     y offset from the diagonal  (odd)
!              naz     z offset from the diagonal  (odd)
!         I/O:
!              in      green_in function (nox,noy,nz,nsux,nsuy) 
!              out     Hessian           (naz,nax,nay,nz,nx,ny) 
!                          
! Description: 
!              Computes the Hessian 
!              Parallelize in the omega dimension
! ---------------------------------------------------------------------------

program L_tL_ng_3D_off
use sep
use omp_lib
use omp_sector_real_off

implicit none

integer                                    :: n1,n2,n3,n4,n5,n6
real                                       :: o1,o2,o3,o4,o5,o6
real                                       :: d1,d2,d3,d4,d5,d6
integer                                    :: esize
integer                                    :: nsux,nsuy,nz,nx,ny,nax,nay,naz,nax2,nay2,naz2,nxd,nyd,ratiox,ratioy,nry,nrx,nsy,nsx
integer                                    :: noffx,noffy
integer                                    :: ix,iy,iz,i1,i2,i3,i4,i5,i6,ixdata,iydata,ioffx,ioffy
real                                       :: minix,miniy,maxix,maxiy,minidx,minidy,maxidx,maxidy,xdata,ydata
complex, dimension(:,:,:,:,:), allocatable :: green,green_input
real,    dimension(:)        , allocatable :: auxi
!real   , dimension(:,:,:,:,:,:,:,:), allocatable :: LtL,auxi2 I took out the 8 dimension because of the Fortran limitation
real   , dimension(:,:,:,:,:,:,:), allocatable :: LtL,auxi2 
integer, external                          :: srite,sreed
logical                                    :: add,marine,logic
integer                                    :: time_array_0(8), time_array_1(8),a1,a2,a3,a4,a5
real                                       :: finish_timet, start_timet

call sep_init()
call set_no_putch()

call from_history('n1',n1); call from_history('d1',d1); call from_history('o1',o1) !off_x
call from_history('n2',n2); call from_history('d2',d2); call from_history('o2',o2) !off_y
call from_history('n3',n3); call from_history('d3',d3); call from_history('o3',o3) 
call from_history('n4',n4); call from_history('d4',d4); call from_history('o4',o4) !z
call from_history('n5',n5); call from_history('d5',d5); call from_history('o5',o5) !sx
call from_history('n6',n6); call from_history('d6',d6); call from_history('o6',o6) !sy
call from_history("esize",esize)

call from_param('nax',nax,11)        ! allways odd 
call from_param('nay',nay,11)        ! allways odd 
call from_param('naz',naz,11)        ! allways odd
call from_param('noffx',noffx,1)     ! subsurface offset
call from_param('noffy',noffy,1)     ! subsurface offset
call from_param('minix',minix,-999999999.)  
call from_param('miniy',miniy,-999999999.)  
call from_param('nx',nx,-999999999)  
call from_param('ny',ny,-999999999)  
call from_param('ratiox',ratiox,1)       ! source to receiver ratio 
call from_param('ratioy',ratioy,1)       ! source to receiver ratio
call from_param('nrx',nrx,1)        
call from_param('nry',nry,1)
call from_param('nsx',nsx,1)        
call from_param('nsy',nsy,1)
call from_param('marine',marine,.false.)  

     
! Parameters calculation out of the loop
if (marine) then 
  minidx=o4+o1
  minidy=o5+o2
  nxd=1.5+((n5-1)*d5+(n1-1)*d1)/d1                
  nyd=1.5+((n6-1)*d6+(n2-1)*d2)/d2                
else
  minidx=o1
  minidy=o2
  nxd=n1
  nyd=n2
endif

if (nx   ==-999999999 ) nx = nxd
if (ny   ==-999999999 ) ny = nyd
if (minix==-999999999.) minix = minidx
if (miniy==-999999999.) miniy = minidy
maxix = minix + (nx-1)*d1
maxiy = miniy + (ny-1)*d2
nsux=n5
nsuy=n6
nz=n4
naz2=(naz-1)/2
nax2=(nax-1)/2
nay2=(nay-1)/2

!check to see if this is the first time we are writing to the file----------------------------------------------------
call from_param("add",add,.false.)
if (.not. add) then
  call to_history('n1',naz,   "outtag");  call to_history('d1',d4,       "outtag")
  call to_history('o1',naz2,  "outtag");  call to_history('label1',"az", "outtag")
  call to_history('n2',nax,   "outtag");  call to_history('d2',d1,       "outtag") 
  call to_history('o2',nax2,  "outtag");  call to_history('label2',"ax", "outtag")
  call to_history('n3',nay,   "outtag");  call to_history('d3',d2,       "outtag") 
  call to_history('o3',nay2,  "outtag");  call to_history('label3',"ay", "outtag")
  call to_history('n4',noffx, "outtag");  call to_history('d4',d1,       "outtag") 
  call to_history('o4',0.,    "outtag");  call to_history('label4',"hx", "outtag")
  call to_history('n5',noffy, "outtag");  call to_history('d5',d2,       "outtag") 
  call to_history('o5',0.,    "outtag");  call to_history('label5',"hy", "outtag")
  call to_history('n6',nz,    "outtag");  call to_history('d6',d4,       "outtag") 
  call to_history('o6',o4,    "outtag");  call to_history('label6',"z",  "outtag")
  call to_history('n7',nx,    "outtag");  call to_history('d7',d1,       "outtag") 
  call to_history('o7',minix, "outtag");  call to_history('label7',"mx", "outtag")
  call to_history('n8',ny,    "outtag");  call to_history('d8',d2,       "outtag") 
  call to_history('o8',miniy, "outtag");  call to_history('label8',"my", "outtag")
  call to_history('n9',noffx, "outtag");  call to_history('d9',d1,       "outtag") 
  call to_history('o9',0.,    "outtag");  call to_history('label9',"hx", "outtag")
  call to_history('n10',noffy,"outtag");  call to_history('d10',d2,      "outtag") 
  call to_history('o10',0.,   "outtag");  call to_history('label10',"hy","outtag")
  call to_history('n11',1,    "outtag");  call to_history('d11',1.,      "outtag") 
  call to_history('o11',0.,   "outtag");  call to_history('label11',"w", "outtag")
  call to_history('esize',esize/2,"outtag")
  do iy=1,ny
    allocate(auxi(naz*nax*nay*noffx*noffy*nz*nx*ny))
    auxi=0.
    if (srite("outtag",auxi,esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2) /= esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2) &
      call seperr("Error initializing Hessian")
    deallocate(auxi)
  enddo
  call auxclose("outtag")
end if
call auxinout("outtag")
!---------------------------------------------------------------------------------------------------------------------
call sep_begin_prog()
!----------------------------------------------------------------------------------------------------------------------
call date_and_time(values=time_array_0)
start_timet = time_array_0 (3) * 86400.+ time_array_0 (5) * 3600  + time_array_0 (6) * 60 &
           + time_array_0 (7) + 0.001 * time_array_0 (8)
!----------------------------------------------------------------------------------------------------------------------
  allocate(green_input(n1,n2,n4,n5,n6),green(nsuy,nsux,nz+2,nx+2,ny+2))
  call sep_read(green_input)
  green=(0.,0.)
  do i1=1,n1
  do i2=1,n2
  do i4=1,n4
  do i5=1,n5
    xdata=o5+(i5-1)*d5+o1+(i1-1)*d1                                                        
    if ((xdata >= minix) .and. (xdata <= maxix)) then
    do i6=1,n6     
      ydata=o6+(i6-1)*d6+o2+(i2-1)*d2
      if ((ydata >= miniy) .and. (ydata <= maxiy)) then
        ixdata=1.5+(xdata-minix)/d1
        iydata=1.5+(ydata-miniy)/d2
        green(i6,i5,i4+1,ixdata+1,iydata+1)=green_input(i1,i2,i4,i5,i6) ! Very slow operation I need to look into this (Transpose outside?)
      endif
    end do
    end if
  end do
  end do
  end do
  end do

  deallocate(green_input)
!----------------------------------------------------------------------------------------------------------------------
  allocate(LtL(naz,nax,nay,noffx,nz,nx,ny),auxi2(naz,nax,nay,noffx,nz,nx,ny)) 
  write(0,*)"allocated memory green",nsux*nsuy*nz*nx*ny*esize 
  write(0,*)"allocated memory LtL"  ,naz*nax*nay*nz*nx*ny*noffx*noffy*esize/2
  write(0,*)"allocated memory auxi2",naz*nax*nay*nz*nx*ny*noffx*noffy*esize/2
  write(0,*)"n1,n2,n3,n4,n5,n6",n1,n2,n3,n4,n5,n6
  write(0,*)"nx,ny,nz,nsux,nsuy",nx,ny,nz,nsux,nsuy
  write(0,*)"d1,d2,d3,d4,d5,d6",d1,d2,d3,d4,d5,d6
  write(0,*)"o1,o2,o3,o4,o5,o6",o1,o2,o3,o4,o5,o6
  write(0,*)"marine,ratiox", marine,ratiox
  write(0,*)"marine,ratioy", marine,ratioy

 
 do ioffx=1,noffx
 do ioffy=1,noffy
    LtL=0.
   !$omp parallel do private(ix)
    do ix=1,nx
    do iy=1,ny
      call omp_do (nx,ny,nz,noffx,noffy,nax,nay,naz,nax2,nay2,naz2,ix,iy,ioffx,ioffy,green,LtL,ratiox,ratioy,nsux,nsuy,nry,nrx,nsy,nsx)
    enddo
    enddo  
   !$omp end parallel do
   !---------------------------------------------------------------------------------------------------------------------
   call sep_open_lock("outtag")  !make sure only one process at a time writes
   call sseek("outtag",(ioffx-1)*(ioffy-1)*esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2,0)
   if (sreed("outtag",auxi2,esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2) /= esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2) &
      call seperr("Error reading data")
   LtL= LtL+auxi2
   call sseek("outtag",0,0)
   call sseek("outtag",(ioffx-1)*(ioffy-1)*esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2,0)
   if (srite("outtag",LtL,auxi2,esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2) /= esize*naz*nax*nay*noffx*noffy*nz*nx*ny/2) &
      call seperr("Error writing data")
   call sep_close_lock("outtag") !allow other processes to write
   !---------------------------------------------------------------------------------------------------------------------
 enddo
 enddo

!----------------------------------------------------------------------------------------------------------------------
call date_and_time(values=time_array_1)
finish_timet = time_array_0 (3) * 86400.+ time_array_1 (5) * 3600  + time_array_1 (6) * 60 &
            + time_array_1 (7) + 0.001 * time_array_1 (8)
write(0,*)'total time', finish_timet - start_timet
!----------------------------------------------------------------------------------------------------------------------

deallocate(green,LtL,auxi2)

call sep_end_prog()

end program 
