module wei_SPig
  !! image up- and down-going wave-fields for Shot Profile migration 
  !!      output is offset domain

  use wei_util
  use wei_external
  use wei_run
  use wei_down
  use wei_flags
  use wei_R
  use wei_type
  use wei_fft
  !---------------
  implicit none
  integer ::o_offx,o_offy, zoffx, zoffy
contains
  !----------------------------------------------------------------  
  !! offset domain Shot-profile imaging
  integer function SPig_init() result(st)

#ifdef DEBUG
    character(len=128), parameter :: unit='SPimag_init'
    if(run%debug) call in(unit)
#endif

    o_offx=down%ahx%o/down%ahx%d
    o_offy=down%ahy%o/down%ahy%d
    zoffy=1+down%ahy%n/2
    zoffx=1+down%ahx%n/2
  st=0
  end function SPig_init
  !----------------------------------------------------------------
  integer function SPigop(model,wfld_u,wfld_d,R) result(st)
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    complex, dimension(:,:,:,:)  , pointer :: model
    type(imag), intent(inout) ::R
    integer   :: j,i,k,it,f1,f2,j1,j2,i3,i4
#ifdef DEBUG
    character(len=128), parameter :: unit='SPimag'
    if(run%depar) call in(unit)
    if(run%verb)write(0,*) 'imwf/model in igop=',cabs(sum(model))
#endif

    freqs:do k=1,down%nws
       !   inline offsets 
          y:do j=1, down%amy%n
             xoffs:do i=1,down%ahx%n
                it=i+o_offx-1
                it=2*it
                f1=-floor((real(it)+0.1)/2.)
                f2=f1+it
                j1=max(1,max(1-f1,1-f2))
                j2=min(down%amx%n,min(down%amx%n-f1,down%amx%n-f2))
                   model(j1:j2,j,i,zoffy)=model(j1:j2,j,i,zoffy) + &
                     wfld_u( j1+f2:j2+f2, j,k)*conjg(wfld_d(j1+f1:j2+f1, j,k))
                if(down%norm) model(j1:j2,j,i,zoffy)=model(j1:j2,j,i,zoffy)/&
                   (wfld_d( j1+f2:j2+f2, j,k)*conjg(wfld_d(j1+f1:j2+f1, j,k)))
             end do xoffs
          end do y

       !!   xline offsets 
       x:do j=1, down%amx%n
          yoffs:do i=1,down%ahy%n
             it=i+o_offy-1
             it=2*it
             f1=-floor((real(it)+0.1)/2.)
             f2=f1+it
             j1=max(1,max(1-f1,1-f2))
             j2=min(down%amy%n,min(down%amy%n-f1,down%amy%n-f2))
                model(j,j1:j2,zoffx,i)= model(j,j1:j2,zoffx,i) + &
                  wfld_u(j,j1+f2:j2+f2, k)*conjg(wfld_d(j, j1+f1:j2+f1, k))
             if(down%norm) model(j,j1:j2,zoffx,i)= model(j,j1:j2,zoffx,i)/ &
               (wfld_d(j,j1+f2:j2+f2, k)*conjg(wfld_d(j, j1+f1:j2+f1, k)))
          end do yoffs
       end do x

       if (down%image_freqs) then
          R%bin(:,:,:,:,1)=model
          model=cmplx(0.,0.) !! zero depth slice before summation over frequency
          if (.not. image_insert(R))&
          call seperr("Single shot image => global image didnt go well")
          do i4=1,size(R%bin_tot,4)
             do i3=1,size(R%bin_tot,3)
                call write_mx_my(R%tag,R%esize,size(R%bin_tot,1),size(R%bin_tot,2),&
                R%bin_tot(:,:,i3,i4,1),down%shotnum/=0)
             end do
          end do
       end if

    end do freqs
#ifdef DEBUG
    if(run%verb)write(0,*) 'imwf/model after loops igop=',cabs(sum(model))
    if(run%depar) call out(unit)
#endif
    st=0
  end function SPigop

  !----------------------------------------------------------------
  integer function SPigop_m(model,wfld_u,wfld_d,R) result(st)
    complex, dimension(:,:,:), pointer :: wfld_u,wfld_d
    complex, dimension(:,:,:,:)  , pointer :: model
    type(imag), intent(inout) ::R
    integer   :: j,i,k,it,f1,f2,j1,j2,i3,i4
#ifdef DEBUG
    character(len=128), parameter :: unit='SPimag_m'
    if(run%depar) call in(unit)
    if(run%verb)write(0,*) 'imwf/model in igop=',cabs(sum(model))
#endif

    freqs:do k=1,down%nws
       !   inline offsets 
          y:do j=1, down%amy%n
             xoffs:do i=1,down%ahx%n
                it=i+o_offx-1
                it=2*it
                f1=-floor((real(it)+0.1)/2.)
                f2=f1+it
                j1=max(1,max(1-f1,1-f2))
                j2=min(down%amx%n,min(down%amx%n-f1,down%amx%n-f2))
                   model(j1:j2,j,i,zoffy)=model(j1:j2,j,i,zoffy) + &
                     wfld_u( j1+f2:j2+f2, j,k)*wfld_u(j1+f1:j2+f1, j,k)
                if(down%norm) model(j1:j2,j,i,zoffy)=model(j1:j2,j,i,zoffy)/&
                   (wfld_d( j1+f2:j2+f2, j,k)*conjg(wfld_d(j1+f1:j2+f1, j,k)))
             end do xoffs
          end do y

       !!   xline offsets 
       x:do j=1, down%amx%n
          yoffs:do i=1,down%ahy%n
             it=i+o_offy-1
             it=2*it
             f1=-floor((real(it)+0.1)/2.)
             f2=f1+it
             j1=max(1,max(1-f1,1-f2))
             j2=min(down%amy%n,min(down%amy%n-f1,down%amy%n-f2))
                model(j,j1:j2,zoffx,i)= model(j,j1:j2,zoffx,i) + &
                  wfld_u(j,j1+f2:j2+f2, k)*wfld_u(j, j1+f1:j2+f1,k)
             if(down%norm) model(j,j1:j2,zoffx,i)= model(j,j1:j2,zoffx,i)/ &
               (wfld_d(j,j1+f2:j2+f2, k)*conjg(wfld_d(j, j1+f1:j2+f1, k)))
          end do yoffs
       end do x

       if (down%image_freqs) then
          R%bin(:,:,:,:,1)=model
          model=cmplx(0.,0.) !! zero depth slice before summation over frequency
          if (.not. image_insert(R))&
          call seperr("Single shot image => global image didnt go well")
          do i4=1,size(R%bin_tot,4)
             do i3=1,size(R%bin_tot,3)
                call write_mx_my(R%tag,R%esize,size(R%bin_tot,1),size(R%bin_tot,2),&
                R%bin_tot(:,:,i3,i4,1),down%shotnum/=0)
             end do
          end do
       end if

    end do freqs
#ifdef DEBUG
    if(run%verb)write(0,*) 'imwf/model after loops igop=',cabs(sum(model))
    if(run%depar) call out(unit)
#endif
    st=0
  end function SPigop_m
!----------------------------------------------------------------------------------------
end module wei_SPig
