! ---------------------------------------------------------------------------
! L_tL_parallel_test6.f90  - by Alejandro Valenciano 
!                        
!           March  2005.
! ---------------------------------------------------------------------------
!       Usage: 
!              L_tL_parallel_test6.x  
!  Parameters: 
!              nxh     x offset to calculate the Hessian (odd)
!              nzh     z offset to calculate the Hessian (odd)
!              tsid    true if receiver at each side of the shot
!         I/O:
!              in      green_in function (nsx,nz,nx) 
!              out     Hessian           (nz,nx_out,nhz,nhx,w) 
!                          
! Description: 
!              Computes the Hessian 
!              Parallelizable in the omega dimension
!
!   n1=offset
!   n2=depth
!   n3=shot_position
!   n4=frequency
!
! ---------------------------------------------------------------------------

program L_tL_para 
use sep

implicit none

integer                                  :: n1,n2,n3,n4
integer                                  :: nsx,nz,nx,nw
real                                     :: o1,o2,o3,o4
real                                     :: d1,d2,d3,d4 
integer                                  :: esize,iz,ix,ihz,ihx,isr,iw,istart,i1
real                                     :: mini,maxi,start
integer                                  :: tsid
complex, dimension(:,:,:)  , allocatable :: green,green_in
complex, dimension(:)      , allocatable :: aaaixz,aaajxz 
real,    dimension(:,:,:,:), allocatable :: LtL,auxi 
complex, dimension(:,:,:,:), allocatable :: MulGreen 
integer, external                        :: srite,sreed
logical                                  :: add

call sep_init()
call set_no_putch()

call from_history('n1',n1); call from_history('d1',d1); call from_history('o1',o1)
call from_history('n2',n2); call from_history('d2',d2); call from_history('o2',o2)
call from_history('n3',n3); call from_history('d3',d3); call from_history('o3',o3)
call from_history('n4',n4); call from_history('d4',d4); call from_history('o4',o4)
call from_history("esize",esize)

call from_param('tsid',tsid,0)       ! 1 if receiver at each side of the shot
if (tsid==1) then                    ! same # receiver at each side of the shot
  mini=o3+o1
else                                 ! receiver at the rigth of the shot 
  mini=o3                            ! assumes o3=0 first offset=0 
endif
maxi=mini+(n3-1)*d3+(n1-1)*d1
nx=1.5+(maxi-mini)/d1                ! new nx
nsx=n3
nz=n2
nw=n4

!check to see if this is the first time we are writing to the file
call from_param("add",add,.false.)

if (.not. add) then
  call to_history('n1',nz,"outtag");call to_history('d1',d2,"outtag");call to_history('o1',o2,"outtag");call to_history('label1',"z","outtag")
  call to_history('n2',nx,"outtag");call to_history('d2',d1,"outtag");call to_history('o2',mini,"outtag");call to_history('label2',"mx","outtag")
  call to_history('n3',nz,"outtag");call to_history('d3',d2,"outtag");call to_history('o3',o2,"outtag");call to_history('label3',"z","outtag")
  call to_history('n4',nx,"outtag");call to_history('d4',d1,"outtag");call to_history('o4',mini,"outtag");call to_history('label4',"mx","outtag")
  call to_history('n5',nw,"outtag");call to_history('d5',d4,"outtag");call to_history('o5',o4,"outtag");call to_history('label5',"w","outtag")
  call to_history('esize',4,"outtag")
  allocate(auxi(nz,nx,nz,nx))
  auxi=0.
  if (srite("outtag",auxi,4*nz*nx*nz*nx) /= 4*nz*nx*nz*nx) &
    call seperr("Error writing data")
  call auxclose("outtag")
  deallocate(auxi)
end if
call auxinout("outtag")

call sep_begin_prog()

write(0,*)"nsx,nz,nx,nw",nsx,nz,nx,nw
write(0,*)"d1,d2,d3,d4",d1,d2,d3,n4
write(0,*)"o1,o2,o3,o4",o1,o2,o3,n4

allocate(green_in(n1,nz,nsx),green(nsx,nz,nx),LtL(nz,nx,nz,nx), &
         MulGreen(nz,nx,nz,nx),aaaixz(nsx),aaajxz(nsx),auxi(nz,nx,nz,nx))
write(0,*)"allocated memory green_in",n1*nz*nsx*8
write(0,*)"allocated memory green",nsx*nz*nx*8
write(0,*)"allocated memory LtL",nz*nx*nz*nx*4
write(0,*)"allocated memory auxi",nz*nx*nz*nx*4
write(0,*)"allocated memory MulGreen",nz*nx*nz*nx*8
write(0,*)"allocated memory aaa",2*nsx*8

write(0,*)"allocated memory",n1*nz*nsx*8+nsx*nz*nx*8+2*nz*nx*nz*nx*4+nz*nx*nz*nx*8+2*nsx*8

do iw=1,nw
   write(0,*)"frequency",iw
   call sep_read(green_in)
   green=(0.,0.)
   do isr=1,nsx
     start=mini+(isr-1)*d3
     istart=1.5+(start-mini)/d1
     do iz=1,nz
     do i1=1,n1
       green(isr,iz,istart+i1-1)=green_in(i1,iz,isr)
     end do
     end do
   end do

   LtL=0.;
   MulGreen=(0.,0.)
   ! Computation of the second Sum
   do ix=1,nx
   do iz=1,nz
     aaaixz(1:nsx)=conjg(green(1:nsx,iz,ix))
     do ihx=1,nx
     do ihz=1,nz
       aaajxz(1:nsx)=green(1:nsx,ihz,ihx)          !  G(h_z,h_x)
       do isr=1,nsx
         MulGreen(ihz,ihx,iz,ix)=MulGreen(ihz,ihx,iz,ix)+aaaixz(isr)*aaajxz(isr)
       enddo
     enddo
     enddo
   enddo
   enddo
   
   ! Computation of the Hessian
   do ix=1,nx
   do iz=1,nz
     aaaixz(1:nsx)=conjg(green(1:nsx,iz,ix))
     do ihx=1,nx
     do ihz=1,nz
        aaajxz(1:nsx)=green(1:nsx,ihz,ihx)
        do isr=1,nsx
         LtL(ihz,ihx,iz,ix)=LtL(ihz,ihx,iz,ix)+ &
                                  real( aaaixz(isr)*aaajxz(isr)*MulGreen(ihz,ihx,iz,ix) )
       enddo
     enddo
     enddo
   enddo
   enddo

   call sep_open_lock("outtag")  !make sure only one process at a time writes

   if (sreed("outtag",auxi,esize*nz*nx*nz*nx/2) /= esize*nz*nx*nz*nx/2) &
     call seperr("Error reading data")
   LtL= LtL+auxi
   call sseek("outtag",0,0)

   if (srite("outtag",LtL,esize*nz*nx*nz*nx/2) /= esize*nz*nx*nz*nx/2) &
      call seperr("Error writing data")

   call sep_close_lock("outtag") !allow other processes to write
enddo

deallocate(green_in,green,LtL,MulGreen,aaaixz,aaajxz,auxi)

call sep_end_prog()

end program L_tL_para
