! ---------------------------------------------------------------------------
!  Pad.f90 - Modified by Alejandro Valenciano based on
!                        Fxcorr.f90 by Alejandro Valenciano
!           March  2003.
! ---------------------------------------------------------------------------
!       Usage: Pad.x
!  Parameters:  
!              in      wavefield (mx,my,w,z,sx) 
!              out     wavefield (x,y,w,z,sx) 
!                          
! Description: Arrange the output of Brad program to work with my imaging condition. 
!              Pad the fast direction to make it image x position 
! ---------------------------------------------------------------------------

program  Pad_mx_sym
        use sep
        implicit none

        integer                                 :: n1,n2,n3,n4,n5,n6,esize
        real                                    :: d1,d2,d3,d4,d5,d6
        real                                    :: o1,o2,o3,o4,o5,o6
        real                                    :: min,max,start 
        integer                                 :: n,i,j,istart
        integer                                 :: block_size,n_block
        real,   dimension(:), allocatable    :: in,out
        integer, external                       :: sreed,sseek_block,srite
        !! hook up SEP init stuff and get input parameters from history file
        call sep_init ()
        call from_history ("n1",n1); call from_history ("d1",d1); call from_history ("o1",o1)
        call from_history ("n2",n2); call from_history ("d2",d2); call from_history ("o2",o2)
        call from_history ("n3",n3); call from_history ("d3",d3); call from_history ("o3",o3)
        call from_history ("n4",n4); call from_history ("d4",d4); call from_history ("o4",o4)
        call from_history ("n5",n5); call from_history ("d5",d5); call from_history ("o5",o5)
        call from_history ("n6",n6); call from_history ("d6",d6); call from_history ("o6",o6)
        call from_history ("esize",esize)
!! calculate the new n (same # receiver at each side of the shot) 
        min=o6+o1
        max=o6+(n6-1)*d6+o1+(n1-1)*d1
        n=1.5+(max-min)/d1

        write(0,*)"min,max,n",min,max,n

      !! output parameters to history file
      call to_history ("n1",n); call to_history ("o1",min)

        !! allocate working space
        allocate (in(n1),out(n))

        block_size=n1*esize
        n_block=0
        do i=1,n6
         !! Positioning
         start=o1+o6+(i-1)*d6
         istart=1.5+(start-min)/d1

        write(0,*)"start,istart,shotx",start,istart,o6+(i-1)*d6

         do j=1,n2*n3*n4*n5
          !! Extract mx column 
          if (sseek_block("in",n_block,block_size,0) /=n_block) &
             call seperr("Error seeking wavefield")
          if (sreed("in",in,block_size) /=block_size) &
             call seperr("Error reading wavefield")
          !! Positioning 
          out=(0.,0.)
          out(istart:istart+n1-1)=in(1:n1)    
          n_block=n_block+1
          !! write output 
          if ( srite ("out",out,esize*n) /= esize*n) write(0,*) "Error writing wavefield"
         end do
        end do

        deallocate (in,out)
end program  Pad_mx_sym
