module wei_SPmig
  !! WEI migration, Shot Profile style
  use wei_util
  use wei_run
   
  use wei_S
  use wei_R
  use wei_D
  use wei_W

  use wei_fft
  use wei_kmap
  use wei_xmap

  use wei_kxmig
  use wei_image

  implicit none

  type(dat),public :: D,D_d
  type(imag),public :: R,M
!  integer     :: time_array_0(8), time_array_1(8)
!  real        :: start_time, finish_time

  integer::d_dpresent,wpresent,w_dpresent,mpresent
!  character(len=128) :: name1,name2
contains
  subroutine SPmig_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
    integer,external,optional ::                    in6,in7,in8,in9
    integer :: st,er

#ifdef DEBUG
    character(len=128), parameter :: unit='mig_init'
#endif
    interface !!-------------------------
       integer function    SLin() result(st)
       end function SLin
       integer function    WCin() result(st)
       end function WCin
       integer function    FKin() result(st)
       end function FKin
       integer function    FXin() result(st)
       end function FXin
       integer function    IGin() result(st)
       end function IGin
    end interface !!---------------------


    !! initialize the main WEI variables
    if(.not. run_init()) &
    call erexit("trouble initializing run_init")
    if(.not. down_init()) &
    call erexit("trouble initializing down_init")
    if(.not. init_distrib_pattern()) &
    call erexit("trouble initializing distribution pattern")


#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    !! initialize objects
    if(run%verb) call message("Initialize data and source")
    if(.not. data_init(D,down%D)) &
    call erexit("trouble initializing D")
!    if (.not.down%passive .or. .not.down%teleseismic)then
    if (.not.down%passive )then
       if(.not. source_init(D_d,down%D_d)) & 
       call erexit("trouble initializing D_d")
    else
       if(.not. aux_data_init(D_d,down%D_d)) &
       call erexit("trouble initializing D_d")
!       call data_copy(D ,D_d)    !! src=recvr for passive
    end if

    if(run%verb) call message("Initialize Slowness")
    if(.not. slow_init(S,down%VS,down%VG))&
    call erexit("trouble initializing slowness")

    if (down%marine) call data_indices_marine(D,D_d)

    if(run%verb) call message( "Initalizing Image ")
    if(.not. image_init(R,down%R)) &
            call erexit("trouble initializing image R")
    if (0.ne.getch("M",'d',mpresent) ) then
       down%mult=.true.
       if(.not. image_init(M,down%M))call seperr('badness at image_init(M')
       if(run%verb) write(0,*) "        *      Writing imagespace multiple model =",down%mult
    end if

    if (0.ne.getch("W",'d',wpresent) )then
       if(down%image_freqs)call seperr("piss off, you cant have everything")
       down%wIOu=.true.
       if(.not. bwfld_init(W  ,down%W  ))call seperr('badness at bwfld_init(W')
       if(run%verb) write(0,*) "        *      Writing wavefield file W.    wIOu =",down%wIOu
    end if
    if (0.ne.getch("W_d",'d',w_dpresent) ) then
       down%wIOd=.true.
       if(down%image_freqs)call seperr("piss off, you cant have everything")
       if(.not. bwfld_init(W_d,down%W_d))call seperr('badness at bwfld_init(W_d')
       if(run%verb) write(0,*) "        *      Writing wavefield file W_d.  wIOd =",down%wIOd
    end if

!! . .  Maping files to tags
    down%mod_    = down%R     !SINGLE SHOT MODEL (REFLECTIVITY)
    if(down%mult)  down%m_ = down%M     !image space SRME model
    down%dat_    = down%D     !DATA IS FREQUENCY SLICES
    down%dat_d_  = down%D_d   !SOURCE DATA IS FREQUENCY SLICES

    if(run%verb) call message( "Initialize Fourier stuff")
    !! convert frequency to radians
      call process_radians()
    !! init FFT
    call fft_init_r(D%bin)  
    call fft_init_s(D_d%bin)  
    !! init K-map
    call kmap_init()
    call xmap_init()

    !! init imaging
    call imaging_init()

    if(run%verb) call message( "Allocating slowness arrays")
    !! allocate slowness arrays for FK/FX ops
    call slow_allocate()

    if(run%verb) call message( "Initializing all operators")
    !! init operators
    st=SLin()     !! FK slownesses init
    st=WCin()     !! WCop init
    st=FKin()     !! FKop init
    st=FXin()     !! FXop init
    st=IGin()     !! FXop init 
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine SPmig_init
  !----------------------------------------------------------------
  subroutine mig_outnull()
  !ZERO OUTPUT IF REQUESTED
#ifdef DEBUG
    character(len=128), parameter :: unit='mig_outnull'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! R<-D
       call     image_assign(R,(0.,0.))
    else             !! R->D
       call      data_assign(D ,(0.,0.))
    end if
    call sep_begin_prog()

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine mig_outnull
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  integer function SPmig(adj,add,mod_,dat_,dat_d_,SLop,WCop,FKop,FXop,IGop,op6,op7,op8,op9)
  logical,            intent(in) :: adj,add
  character(len=128), intent(in) :: mod_,dat_,dat_d_
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb,isx,isy
    integer :: st

#ifdef DEBUG
    character(len=128), parameter :: unit='SPmig'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif
#ifdef TIME
  call date_and_time(values=time_array_0)
  start_time = time_array_0 (5) * 3600 + time_array_0 (6) * 60 &
               + time_array_0 (7) + 0.001 * time_array_0 (8)
#endif

    !! reset tag variables
    call down_reset2(adj,add,mod_,dat_,dat_d_)
!    if(run%verb) call process_follow() !! where am I?
    
    !! accomodate tag name changes
    R%tag     = trim(down%mod_)
    D%tag     = trim(down%dat_)
    D_d%tag   = trim(down%dat_d_)

    !! setup the adjoint sign
    if(down%adj) then
       down%opsign = goT2B
    else
       down%opsign = goB2T
    end if

    !! zero the output if add is off
!    if(.not. down%add) call mig_outnull() !uneseccary, zeroed during initialize

    down%shotnum=0

    shoty: do isy=1,down%asy%n
    shotx: do isx=1,down%asx%n
       if (run%verb)then
          call line()
          write(0,'(a,i5,a,i5)')"         *  Shotnum",&
          down%shotnum+1,' of',down%asy%n*down%asx%n
       end if
       if(down%marine) then
          call data_position(D,isx,isy) 
          call xmap_op()
       else
          call data_indices_land(D_d,isx,isy)
       end if
       call data_position_report()
       if( .not.down%passive)&
       call source_report(D_d)
#ifdef TIME
  if (isx+isy.eq.2) then
    call date_and_time(values=time_array_1)
    finish_time = time_array_1 (5) * 3600 + time_array_1 (6) * 60 &
                + time_array_1 (7) + 0.001 * time_array_1 (8)
  endif
#endif 
       SPmig=SPmigof(isx,isy,SLop,WCop,FKop,FXop,IGop)

       if(run%verb)write(0,'(a,i4)') '         *       Finished migrating shot',down%shotnum +1

       down%shotnum=down%shotnum+1
			 if (.not. down%teleseismic) &
  		 st=sseek(D_d%tag,0,0)               !! rewind to beginning of source
    end do shotx
    end do shoty

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    SPmig=OK
    
	write(0,*)'TERMINE SPMIG OK'
  end function SPmig

!--------------------------------------------------------------------
  integer function SPmigof(isx,isy,SLop,WCop,FKop,FXop,IGop)
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izb,iwb,position
    integer :: st,ierr,isx,isy
#ifdef DEBUG
    character(len=128), parameter :: unit='migof'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
    var%dzb=down%az%d

    if(down%adj) then !! MIGRATION
#ifdef DEBUG
       if(run%debug) call in('migration')
#endif 
       wbA: do iwb=1,down%nwb
          if(run%verb)write(0,'(a,i2,a,i2)') '         *       Frequency block ',iwb,' of',down%nwb
          position=down%shotnum*down%nwb + (iwb-1)

          if(.not. data_read(D,position)) &
          call seperr("trouble reading D data") !! read data at block top
          if (.not.down%passive ) then !!D_d already pointed to D
             if(.not. data_read(D_d,iwb-1)) & !!read the source file
             call seperr("trouble reading D_d data")
          else
             call data_copy(D,D_d)
          end if

          zbA: do izb=1,down%nzb       
             call setup_var_block(iwb,izb)
             R%bin=0.;
             call slowness_read(S,izb) !! read a slowness block
             st = SLop() !! setup sS, sR, s(z) for working space
             down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
             if(down%wIOu .or. down%wIOd) then           !! DC one block
                call wmigWW(izb,iwb,WCop,FKop,FXop,IGop) !! write W
             else
                call wmigRD(izb,iwb,down%nzs,WCop,FKop,FXop,IGop) !! trash W
             end if
             if (.not. down%image_freqs) then
                if (.not. image_insert(R))&
                call seperr("Single shot image => global image didnt go well")
                call image_write(R,izb,.false.)
             end if
          end do zbA
       end do wbA

       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('migration')
#endif
    else         !! MODELING
#ifdef DEBUG
       if(run%debug) call in('modeling')
#endif
       !----------------------------------------------------------

#ifdef DEBUG
       if(run%debug) call out('modeling')
#endif

       !! no modeling operator yet

    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    SPmigof=OK
  end function SPmigof 
  !---------------------------------------------------------------
  !! don't write the wavefield
  subroutine wmigRD(izb,iwb,nzs,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop 
    integer :: izs,iws,nzs,a
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigRD'
    if(run%debug) call in(unit)
#endif
    if(down%adj) then !! MIGRATION
       if (down%image_freqs) a=sseek(R%tag,0,0)
       zsA: do izs=1,nzs          !LOOP OVER z STEPS
          if(izs.ge.down%firstz) then !! limit file size when min. target depth known
             call SPimg(izs,R,D%bin,D_d%bin,IGop)
!             if (down%mult) call SPimg(izs,M,D%bin,D_d%bin,IGop_m)
          end if
          if (mod(izs,down%az%n/10)==0) call depth_report(izs,izb,iwb)
          if (izs==down%nzs) return !! no need to DC further than the model space
          wsA: do iws=1,down%nws             !LOOP OVER w STEPS
             call kxmig(D%bin,D_d%bin,iws,izs,WCop,FKop,FXop)
          end do wsA
       end do zsA
    else         !! MODELLING
       call erexit('Shot profile modeling not yet implemented.')
!       wsF: do iws=1,down%nws             !LOOP OVER w STEPS
!          zsF: do izs=down%nzs,1,-1       !LOOP OVER z STEPS
             !!need to write imgSPadj
             !!until then, do nothing
!             call kxmig(bW%bin,bW_d%bin,iws,izs,WCop,FKop,FXop)
!             call imgSPadj(izs,iws,R%bin,bW%bin,IGop)
!          end do zsF
!       end do wsF
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigRD
  !----------------------------------------------------------------
  !! write the wavefield
  subroutine wmigWW(izb,iwb,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izs,iws
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigWW'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! MIGRATION
       zsA: do izs=1,down%nzs
          if(izs.ge.down%firstzW) then !! limit W file size when min. target depth known
             if (down%wIOu)then
                if (.not. wavefield_write(W  ,D%bin  ,iwb,izb,izs-down%firstzW+1)) &
                call seperr('wavefield_write(W ==badness')
             end if
             if (down%wIOd)then
                if (.not. wavefield_write(W_d,D_d%bin,iwb,izb,izs-down%firstzW+1)) &
                call seperr('wavefield_write(W_d ==badness')
             end if
          endif
          if(izs.ge.down%firstz) then !! limit R file size when min. target depth known
             call SPimg(izs,R,D%bin,D_d%bin,IGop)
             if (down%mult) call SPimg(izs,M,D%bin,D_d%bin,IGop)
          end if

          if (mod(izs,down%az%n/10)==0) call depth_report(izs,izb,iwb)
          if (izs==down%nzs) return

          wsA: do iws=1,down%nws
             call kxmig(D%bin,D_d%bin,iws,izs,WCop,FKop,FXop)
          end do wsA
        end do zsA
    else         !! MODELLING
       call seperr("SP modelling not implented")
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigWW
  !----------------------------------------------------------------

subroutine depth_report(izs,izb,iwb)
integer :: izs,izb,iwb
integer :: pcent,current,total

pcent=(1. + 1.0*(izs+down%nzs*(izb-1))/down%az%n*100/(real(down%nwb)-(iwb-1)))
current=izs+down%nzs*(izb-1)
total=down%nzs*down%nzb
if(run%verb) then
   write(0,'(a,i5,a,i5,a,i2,a)')'         *  Finished depth number',&
      current,' of',total,' (',pcent,'%)'
end if
end subroutine depth_report
!----------------------------------------------------------------
end module wei_SPmig
