program Sparse_MxV_inv_off
use sep
use dottest 
use sparse_matmult 
use cgstep_mod
use solver_mod_mio
use scaled_identity
use sparse_mapping_off

implicit none
real, dimension (:),     allocatable :: m_inv,m_mig,xo,wto
real, dimension (:,:),   allocatable :: remov,mmov
real, dimension (:,:,:,:,:,:), pointer   :: Hcube
real, dimension (:,:)    , pointer   :: Hsp
real, dimension (2)                  :: dot1,dot2
integer                              :: stat
logical                              :: adj,add
integer                              :: niter,na,np  ! iterations
real                                 :: eps,alpha ! scaling parameter
real, dimension(:),      pointer     :: err       ! residual error per iteration
real, dimension(:),      pointer     :: res       ! final residual error
integer                              :: n1,n2,n3,n4,n5,n6 ! nz nx (nhz nhx allways odd)
real                                 :: o1,o2,o3,o4,o5,o6 
real                                 :: d1,d2,d3,d4,d5,d6
integer                              :: esize
integer                              :: i1,i2,i3,i4,ia,cont

call sep_init()

call from_history('n1',n1); call from_history('d1',d1); call from_history('o1',o1)
call from_history('n2',n2); call from_history('d2',d2); call from_history('o2',o2)
call from_history('n3',n3); call from_history('d3',d3); call from_history('o3',o3)
call from_history('n4',n4); call from_history('d4',d4); call from_history('o4',o4)
call from_history('n5',n4); call from_history('d5',d4); call from_history('o5',o4)
call from_history('n6',n4); call from_history('d6',d4); call from_history('o6',o4)
call from_history("esize",esize)

call from_param('eps',eps,.00001)
call from_param('niter',niter,10)
call from_param('na',na,1)
call from_param('alpha',alpha,1.)

call sp_cont (n1,n2,n3,n4,n5,n6,cont)

allocate(Hcube(n1,n2,n3,n4,n5,n6),Hsp(cont,3))

call sep_read(Hcube)                 ! Sep format Hessian matrix

call sp_mapp (Hcube,Hsp,n1,n2,n3,n4,n5,n6) ! Mapp Heesian matrix from Sep to sparse representation

deallocate(Hcube)

allocate(m_inv(n1*n2*n3),m_mig(n1*n2*n3),xo(n1*n2*n3),wto(n1*n2*n3),res(n1*n2*n3))
if( exist_file("remov")) allocate(remov(n1*n2*n3,niter))
if( exist_file("mmov")) allocate(mmov(n1*n2*n3,niter))

call sep_read(m_mig,"mig")  ! Target migration

call smatmult_init (Hsp)
!call dot_test(smatmult_lop,n1*n2,n1*n2,dot1,dot2)
!stat = smatmult_lop(.false.,.false.,m_mig, m_inv)
!write(0,*)'m_mig',m_mig


do ia=1,na
  xo=m_mig
  if (ia==1) then
   wto=1. 
  else 
   wto=1./sqrt(sqrt(1.+((res)/alpha)**2))
  endif

  write(0,*)"wto",maxval(wto),minval(wto)
  call solver_reg(smatmult_lop,cgstep,identity_exec,n1*n2*n3, m_inv, m_mig,niter,eps &
                ,x0=xo,res=res,verb=.true.,wt=wto,xmov=mmov,rmov=remov)
  np=niter*na
  if( exist_file("remov")) then
     call sep_write( remov, "remov")
     call to_history ("n1",n1,"remov")
     call to_history ("n2",n2,"remov")
     call to_history ("n3",n3,"remov")
     call to_history ("n4",np,"remov")
     call to_history ("o1",o1,"remov")
     call to_history ("d1",d1,"remov")
     call to_history ("o2",o2,"remov")
     call to_history ("d2",d2,"remov")
  end if
  if( exist_file("mmov")) then
     call sep_write( mmov, "mmov")
     call to_history ("n1",n1,"mmov")
     call to_history ("n2",n2,"mmov")
     call to_history ("n3",n3,"mmov")
     call to_history ("n4",np,"mmov")
     call to_history ("o1",o1,"mmov")
     call to_history ("d1",d1,"mmov")
     call to_history ("o2",o2,"mmov")
     call to_history ("d2",d2,"mmov")
  end if
  if( exist_file("wmov")) then
     call sep_write( wto, "wmov")
     call to_history ("n1",n1,"wmov")
     call to_history ("n2",n2,"wmov")
     call to_history ("n3",n3,"wmov")
     call to_history ("n4",na,"wmov")
     call to_history ("o1",o1,"wmov")
     call to_history ("d1",d1,"wmov")
     call to_history ("o2",o2,"wmov")
     call to_history ("d2",d2,"wmov")
  end if
end do

call cgstep_close()

call sep_write(m_inv)

call to_history('n1',n1); call to_history('d1',d1); call to_history('o1',o1) ; call to_history('label1',"z" )
call to_history('n2',n2); call to_history('d2',d2); call to_history('o2',o2) ; call to_history('label2',"mx")
call to_history('n3',n3); call to_history('d3',d3); call to_history('o3',o3) ; call to_history('label3',"h" )
call to_history('n4',1); call to_history('d4',1); call to_history('o4',0.) ; call to_history('label4'," ")
call to_history('n5',1); call to_history('d5',1); call to_history('o5',0.) ; call to_history('label5'," ")
call to_history('n6',1); call to_history('d6',1); call to_history('o6',0.) ; call to_history('label6'," ")


if( exist_file("remov")) deallocate(remov)
if( exist_file("mmov")) deallocate(mmov)
deallocate(m_inv,m_mig,Hsp,xo,wto)


end program Sparse_MxV_inv_off
