program Sparse_filtering
use sep
use dottest 
use cgstep_mod
use solver_mod_mio
use scaled_identity

implicit none
real, dimension (:),     allocatable :: data, model,xo,wto
real, dimension (:,:),   allocatable :: remov,mmov
real, dimension (2)                  :: dot1,dot2
integer                              :: stat
logical                              :: adj,add
integer                              :: niter,na,np  ! iterations
real                                 :: eps,alpha ! scaling parameter
real, dimension(:),      pointer     :: err       ! residual error per iteration
real, dimension(:),      pointer     :: res       ! final residual error
integer                              :: n1,n2
real                                 :: o1,o2 
real                                 :: d1,d2
integer                              :: esize
integer                              :: i1,i2,ia

call sep_init()

call from_history('n1',n1); call from_history('d1',d1); call from_history('o1',o1)
call from_history('n2',n2); call from_history('d2',d2); call from_history('o2',o2)
call from_history("esize",esize)

call from_param('eps',eps,.00001)
call from_param('niter',niter,10)
call from_param('na',na,1)
call from_param('alpha',alpha,1.)
allocate(data(n1*n2),model(n1*n2),xo(n1*n2),wto(n1*n2),res(n1*n2))
call sep_read(data)                 



if( exist_file("remov")) allocate(remov(n1*n2,niter))
if( exist_file("mmov")) allocate(mmov(n1*n2,niter))

call dot_test(identity_exec,n1*n2,n1*n2,dot1,dot2)

write(0,*)'data',maxval(data)
do ia=1,na
  xo=data
  if (ia==1) then
   wto=1. 
  else 
   wto=1./sqrt(1.+((model)/alpha)**2)
  endif

  write(0,*)"wto",maxval(wto),minval(wto)
  call scaled_identity_init( wto,n1*n2,ia )
  call solver_reg(identity_exec,cgstep,scaled_identity_exec,n1*n2, model, data,niter,eps &
                ,res=res,verb=.true.,xmov=mmov,rmov=remov)

!  call solver(identity_exec,cgstep, model, data,niter,verb=.true.) 
  np=niter*na
  if( exist_file("remov")) then
     call sep_write( remov, "remov")
     call to_history ("n1",n1,"remov")
     call to_history ("n2",n2,"remov")
     call to_history ("n3",np,"remov")
     call to_history ("o1",o1,"remov")
     call to_history ("d1",d1,"remov")
     call to_history ("o2",o2,"remov")
     call to_history ("d2",d2,"remov")
  end if
  if( exist_file("mmov")) then
     call sep_write( mmov, "mmov")
     call to_history ("n1",n1,"mmov")
     call to_history ("n2",n2,"mmov")
     call to_history ("n3",np,"mmov")
     call to_history ("o1",o1,"mmov")
     call to_history ("d1",d1,"mmov")
     call to_history ("o2",o2,"mmov")
     call to_history ("d2",d2,"mmov")
  end if
  if( exist_file("wmov")) then
     call sep_write( wto, "wmov")
     call to_history ("n1",n1,"wmov")
     call to_history ("n2",n2,"wmov")
     call to_history ("n3",na,"wmov")
     call to_history ("o1",o1,"wmov")
     call to_history ("d1",d1,"wmov")
     call to_history ("o2",o2,"wmov")
     call to_history ("d2",d2,"wmov")
  end if

end do


call cgstep_close()

call sep_write(model)

call to_history('n1',n1); call to_history('d1',d1); call to_history('o1',o1) ; call to_history('label1',"z" )
call to_history('n2',n2); call to_history('d2',d2); call to_history('o2',o2) ; call to_history('label2',"h")


if( exist_file("remov")) deallocate(remov)
if( exist_file("mmov")) deallocate(mmov)
deallocate(data,model,xo,wto,res)


end program Sparse_filtering
