!HANDLE FFTS
module wei_fft
  use wei_flags
  use wei_run
  use wei_S
  implicit none
  integer*8,pointer,private,save :: splanF(:),splanI(:)!,extraF(:),extraI(:)
  integer*8,pointer,private,save :: rplanF(:),rplanI(:)!,extraF(:),extraI(:)
  real, private,save :: in12
contains
  !----------------------------------------------------------------
  !INITIALIZE THE FFT
  subroutine fft_init_r(wfld)
    complex :: wfld(:,:,:)
    integer :: i,dims(2)
#ifdef DEBUG
    character(len=128), parameter :: unit='fft_init'
    if(run%depar) call in(unit)
#endif
    dims=(/size(wfld,1),size(wfld,2)/)
    in12=sqrt(1./real(product(dims)))

    allocate(rplanF(size(wfld,3)))
    allocate(rplanI(size(wfld,3)))

    do i=1,size(wfld,3)!!output writes over input
       call sfftw_plan_dft_2d(rplanF(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),FFTW_FORWARD,FFTW_PATIENT)
       call sfftw_plan_dft_2d(rplanI(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),FFTW_BACKWARD,FFTW_PATIENT)
    end do

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine fft_init_r
!----------------------------------------------------------------
  subroutine fft_init_s(wfld)
    complex :: wfld(:,:,:)
    integer :: i,dims(2)
#ifdef DEBUG
    character(len=128), parameter :: unit='fft_init'
    if(run%depar) call in(unit)
#endif
    dims=(/size(wfld,1),size(wfld,2)/)

    allocate(splanF(size(wfld,3)))
    allocate(splanI(size(wfld,3)))

    do i=1,size(wfld,3)!!output writes over input
       call sfftw_plan_dft_2d(splanF(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),ffTW_FORWARD,fftW_PATIENT)
       call sfftw_plan_dft_2d(splanI(i),dims(1),dims(2),&
       wfld(1,1,i),wfld(1,1,i),ffTW_BACKWARD,fftW_PATIENT)
    end do

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine fft_init_s

  !----------------------------------------------------------------
  !! forward FFT
  subroutine fft(iws,data)
    integer, intent(in) :: iws
    complex, pointer, dimension(:,:,:) :: data
#ifdef DEBUG
    character(len=128), parameter :: unit='fft'
    if(run%depar) call in(unit)
#endif
    if(slo%source) then
       call sfftw_execute(splanF(iws))
    else
       call sfftw_execute(rplanF(iws))
    end if
    data(:,:,iws)=data(:,:,iws)*in12

 end subroutine


  !----------------------------------------------------------------
  !! inverse FFT
  subroutine ift(iws,data)
    integer, intent(in) :: iws
    complex, dimension(:,:,:), pointer :: data
#ifdef DEBUG
    character(len=128), parameter :: unit='ift'
    if(run%depar) call in(unit)
#endif
    if(slo%source) then
       call sfftw_execute(splanI(iws))
    else
       call sfftw_execute(rplanI(iws))
    end if
   data(:,:,iws)=data(:,:,iws)*in12

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine ift

!  !----------------------------------------------------------------
!  !! forward extra FFT
!  subroutine fft_e(ith,ipart,data)
!    integer, intent(in) :: ipart
!    complex, dimension(:,:,:) :: data
!#ifdef DEBUG
!    character(len=128), parameter :: unit='fft_e'
!    if(run%depar) call in(unit)
!#endif
!    iplan=ith+(ipart-1)*my_nomp()
!    
!  data(:,:,ith)=data(:,:,ith)*in12
! end subroutine fft_e
!
!
!  !----------------------------------------------------------------
!  !! inverse FFT
!  subroutine ift_e(ipart,data)
!    integer, intent(in) ::  ipart
!    complex, dimension(:,:,:), pointer :: data
!#ifdef DEBUG
!    character(len=128), parameter :: unit='ifte_e'
!    if(run%depar) call in(unit)
!#endif
!
!    iplan=ith+(ipart-1)*my_nomp()
!    call sfftw_execute(extraI(iplan))
!   data(:,:,ith)=data(:,:,ith)*in12
!#ifdef DEBUG
!    if(run%depar) call out(unit)
!#endif
!  end subroutine ift_e

end module wei_fft
