module wei_hist
  use wei_util
  use wei_run
  !---------------
  implicit none

  real, dimension(:), pointer, private :: histo
  integer, private :: nhist  !! # of elements in histogram

contains
  !----------------------------------------------------------------  
  subroutine hist_init()
    integer :: er
    integer :: mpierr

    !! # of elements in histogram
    if(run%m_ith==0) call from_param("nhist" ,nhist,10)
    if(run%m_nth >0) call MPI_BCAST(nhist, 1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

    allocate(histo(nhist),stat=er)
    if(er/=0) call seperr('cant allocate histo')
    
  end subroutine hist_init
  !----------------------------------------------------------------
  !! return the most popular value of a vector
  real function hist(izs,vector,ivmin,ivmax) result(popular)
    integer, intent(in) :: izs
    real, dimension(:), pointer :: vector
    integer,         intent(in) :: ivmin, ivmax

    real    :: vmin,vmax,dv
    integer :: iv
    integer :: ihist, ihistmax
    real    :: histmax

    vmin = minval(vector(ivmin:ivmax))
    vmax = maxval(vector(ivmin:ivmax))
    dv = (vmax-vmin)/nhist !! bin size

    !! create the histogram
    histo=0
    do iv=ivmin,ivmax
       ihist = 1+floor( (vector(iv)-vmin)/dv )
       if(ihist>nhist .or. ihist<1) cycle
       histo(ihist) = histo(ihist) + 1
    end do

    !! find the most populated bin in the histogram
    ihistmax=1
    histmax=histo(ihistmax)
    do ihist=1,nhist
       if(histo(ihist)>histmax) then
          ihistmax = ihist
          histmax = histo(ihistmax)
       end if
    end do

    popular = vmin + (ihistmax-.5)*dv
  end function hist
  !----------------------------------------------------------------
end module wei_hist











