module wei_kxmig
  !! fk-fx migration
  !! receive wfld at (izs,iws)
  !! return  wfld downward/upward continued
  !! needs Fop, FXop

  use wei_util
  use wei_run
  use wei_external
  use wei_down
  use wei_S
  use wei_fft
  use wei_kmap
  !---------------
  implicit none


  !! temporary interpolation arrays
  complex, dimension(:,:), allocatable, private :: tmp1,tmp2

contains
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine kxmig(wfld_r,wfld_s,iws,izs,KXMIGop,FKop,FXop)
    complex, dimension(:,:,:), pointer         :: wfld_s,wfld_r
    integer,external,optional ::    FKop,FXop
    integer :: izs,iws
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='kxmig'
#endif
    interface !!-------------------------
       integer function KXMIGop(wfld_r,wfld_s,iws,izs,FKop,FXop) result(st)
         complex, dimension(:,:,:), pointer ::wfld_s,wfld_r
         integer, intent(in)       :: iws,izs
         integer,external,optional :: FKop,FXop
       end function KXMIGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    st = KXMIGop(wfld_r,wfld_s,iws,izs,FKop,FXop)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine kxmig

 
  !----------------------------------------------------------------

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function kxmigN_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='kxmigN_init'
    if(run%debug) call in(unit)
#endif

      !INITIALIZE TEMPORARY SPACE FOR WAVEFIELDS 
      call init_wave_temps()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function kxmigN_init

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function kxmigN_DSR(wfld_rec,wfld_src,iws,izs,FKop,FXop) result(st)
    complex, dimension(:,:,:), pointer :: wfld_src,wfld_rec
    integer, intent(in)       :: iws,izs
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='kxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk
         complex, dimension(:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk
         complex, dimension(:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

      !!D.C. SOURCE WAVEFIELD
      call sellect_slow_comp(.true.) !!selects slowness from source field
      st=kxmigN_SSR(wfld_src,iws,izs,Fkop,Fxop)
      !!D.C. RECEIVER WAVEFIELD
      call sellect_slow_comp(.false.)!!selects slowness from recvr field
      st=kxmigN_SSR(wfld_rec,iws,izs,Fkop,Fxop)

    end function kxmigN_DSR
!----------------------------------------------------------------
  !! N reference slownesses
    integer function kxmigN_SSR(wfld,iws,izs,FKop,FXop) result(st)
      complex, dimension(:,:,:), pointer     :: wfld
      integer, intent(in)                    :: iws,izs
      integer                                :: ifk,isource,ierr
#ifdef DEBUG
      character(len=128), parameter :: unit='kxmigN'
#endif
      interface !!----------------------------------------------
         integer function FKop(iws,izs,ifk,wfld) result(st)
           integer, intent(in) :: iws,izs,ifk
           complex, dimension(:,:,:), pointer :: wfld
         end function FKop
         integer function FXop(iws,izs,ifk,wfld) result(st)
           integer, intent(in) :: iws,izs,ifk
           complex, dimension(:,:,:), pointer :: wfld
         end function FXop
      end interface !!------------------------------------------
#ifdef DEBUG
      if(run%depar) call in(unit)
#endif
      if(down%adj) then !! MIGRATION {
         !----------------------------------------------------------------       
         call store_wavefield(wfld,iws) !STORE THE CURRENT WAVEFIELD in tmp2
         do ifk=1,slo%nrs(izs) !{
            call grab_part_wvfld(wfld,ifk,iws,izs) !GRAB A PORTION OF THE WFLD
            st   = FXop(iws,izs,ifk,wfld) !! f-x operator
            call fft(iws,wfld)
            st   = FKop(iws,izs,ifk,wfld) !! f-k operator
            call add_wavefield(wfld,iws) !SUM OUR EFFORTS TOGETHER into tmp1
         end do !}
         call grab_wavefield(wfld,iws) !copy summed up field from tmp1
         if (.not.down%antion) call ift(iws,wfld) 
         !----------------------------------------------------------------
         !}
      else             !! MODELLING {
         !----------------------------------------------------------------     
         call seperr("no. no. no. Tell Alejandro to get to work.")
!         call sellect_slow_comp(isource==1)
!         call store_wavefield(wfld,iws)
!         do ifk=1,slo%nrs(izs) !{
!            call grab_wavefield(wfld,iws)
!            call ift(iws,wfld)
!            st   = FKop(iws,izs,ifk,wfld) !! f-k op
!            call fft(iws,wfld)
!            st   = FXop(iws,izs,ifk,wfld) !! f-x op
!            call store_part_wavefield(wfld,ifk,iws,izs) 
!         end do !}
!         call grab_wavefield(wfld,iws)
!         !----------------------------------------------------------------
      end if!}
#ifdef DEBUG
      if(run%depar) call out(unit)
#endif
      st=OK
    end function kxmigN_SSR
  !----------------------------------------------------------------  
!! multiple fk ops
!!
!!       D(z)
!!       |
!!       v
!!       T1
!!
!!    D1 D2 D3 ...
!!
!!    S1 S2 S3 ...
!!     \ | /
!!       T2
!!       |
!!       v
!!       D(z+dz)
!---------------------------------------------
!  Initialize storage for wavefields
subroutine init_wave_temps() 
 integer ::er

   !! allocate tmp arrays for interpolation of multiple fk ops
   allocate(tmp1(down%amx%n,down%amy%n),stat=er)
   if(er/=0) call seperr('cant allocate tmp1')
   allocate(tmp2(down%amx%n,down%amy%n),stat=er)
   if(er/=0) call seperr('cant allocate tmp2')

 end subroutine init_wave_temps

!----------------------------------------------------------------
subroutine store_wavefield(wfld,iws)
 complex :: wfld(:,:,:)
 integer :: iws

 tmp2(:,:) = wfld(:,:,iws)
 tmp1=0.

end subroutine store_wavefield
!----------------------------------------------------------------
subroutine grab_part_wvfld(wfld,ifk,iws,izs)
  integer :: iws,ifk,ierr,izs
  complex :: wfld(:,:,:)

  wfld(:,:,iws)=cmplx(0.,0.)
  where(slo%si(:,:,izs) == ifk)
     wfld(:,:,iws) = tmp2(:,:)
  end where

end subroutine grab_part_wvfld
!------------------------------------------------------
subroutine add_wavefield(wfld,iws)
complex :: wfld(:,:,:)
integer :: iws

  tmp1(:,:) = tmp1(:,:) +  wfld(:,:,iws)

end subroutine add_wavefield

!----------------------------------------------------------------
subroutine grab_wavefield(wfld,iws)
  complex :: wfld(:,:,:)
!  logical :: first
  integer :: iws
!  integer :: iws,i1,i2,ierr

  wfld(:,:,iws) = tmp1(:,:)
end subroutine grab_wavefield

!----------------------------------------------------------------
subroutine store_part_wavefield(wfld,ifk,iws,izs)
  complex wfld(:,:,:)
  integer :: ifk,iws,izs

  where(slo%si(:,:,izs) == ifk)
     tmp1(:,:) = tmp1(:,:) + wfld(:,:,iws)
  end where
end subroutine store_part_wavefield
!----------------------------------------------------------------
end module wei_kxmig




