module wei_run
  use wei_external
  use wei_util
  implicit none

  type trun
     logical            :: verb    !! verbose
     logical            :: debug   !! Debug
     logical            :: depar   !! debug in the parallel regions
     logical            :: restart !! whether or not we are restarting the job
     logical            :: omp,mpi !! whether or not this is an omp/mpi job 
     logical            :: share_image !! share image file
     integer            :: io_unit !! unit to write out verb info 
     integer            :: m_ith!! mpi thread number
     integer            :: m_nth !! number of mpi threads
     integer            :: o_nth !! number of omp threads
     character(len=4)   :: tag_num
     character(len=128) :: operator
     character(len=128) :: operation,file_out,local_path
     character(len=9)                      :: mach
  end type trun

   type(trun) :: run

  contains
!----------------------------------------------------------------
!
!
! Function for initializing the run enviro
!
!

  logical function run_init()
    character(len=128) :: operator,operation

#ifdef DEBUG
    character(len=128), parameter :: unit='run_mpi_parallel'
    if(run%debug) call in(unit)
#endif DEBUG
  
    call from_param('verb', run%verb, F)
    call from_param('debug',run%debug,F)
    call from_param('restart',run%restart,F)
    call from_param('depar',run%depar,T)
    if(      run%debug) run%verb =T
    if(.not. run%debug) run%depar=T

    run_init=F
    call from_param('operator',operator,'.....')
    run%operator = operator
    call from_param('operation',operation,'dottest...')
    run%operation = operation
    call from_param("io_file",run%file_out,"NONE")
    if(run%file_out(1:4)=="NONE") then
       run%io_unit=0
    else 
       run%io_unit=6
       open(run%io_unit,file=run%file_out)
    end if

    !Initialize parallel environment for run
    if(.not. run_parallel()) then
       write(0,*) "trouble initializing the run_parallel"
       return
    end if
   run_init=T

   if(run%verb) call run_report()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function run_init



!----------------------------------------------------------------
!
!  Subroutine for initializing parralel run (omp and mpi)
!
!
!
!
  logical function run_parallel()
#ifdef DEBUG
    character(len=128), parameter :: unit='run_parallel'
    if(run%debug) call in(unit)
#endif
 
    run_parallel=F

    if(.not. run_omp_parallel()) then
      write(0,*) "trouble initializing omp enviro"
      return
    end if
    if(.not. run_mpi_parallel()) then
      write(0,*) "trouble initializing mpi enviro"
      return
    end if
    
    run_parallel=T
#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function run_parallel
    

!----------------------------------------------------------------
!
!
! Subroutine to initialize OMP run environment
!
! Sets number of omp threads
!
!
!
!

  logical function run_omp_parallel()
#ifdef DEBUG
    character(len=128), parameter :: unit='omp_run_parallel'
    if(run%debug) call in(unit)
#endif

#ifdef SEP_OMP
    !$OMP PARALLEL
    run%o_nth = omp_get_num_threads()
    !$OMP END PARALLEL
    run%omp=T
#else
    run%o_nth=1
    run%omp=F
#endif

   run_omp_parallel=.true.


#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function  run_omp_parallel

!----------------------------------------------------------------
!
! Function for setting up mpi parallel enviro for mpi
!
! Turns off verb, debug except for the master thread
! Sets up the local datapath
! Figues out the extra to add for each thread
! records the number of threads and this threads num
!
  logical function run_mpi_parallel()
    character(len=128) :: temp_ch
    integer :: ierr,ilen,iverb,igrab,i
#ifdef DEBUG
    character(len=128), parameter :: unit='run_mpi_parallel'
    if(run%debug) call in(unit)
#endif

    run_mpi_parallel=F

    run%share_image=.false.
    run%m_nth=1
    run%m_ith=0

    !TURN OFF VERB AND DEBUG FOR ALL BUT ZERO MPI THREAD
    if(run%m_ith /= 0) then
       run%verb=F
       run%debug=F
    end if


    run%mpi=F


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif

    run_mpi_parallel=T

  end function run_mpi_parallel


  subroutine run_report()
    write(run%io_unit,*) '        WEI*********************************************'
    write(run%io_unit,*) "        *         operator=",trim(run%operator(1:5))," operation=",trim(run%operation(1:9))
    write(run%io_unit,*) "        *         verbose=",run%verb," restart=",run%restart
    write(run%io_unit,*) "        *         debug=",run%debug," para debug", run%depar

  end subroutine run_report
end module wei_run
