module wei_wem
  !! Wave-Equation propagation
  !! full 3-D prestack SSR

  use wei_util
  use wei_down
  use wei_external
  use wei_var
  use wei_run
  use wei_S
  use wei_kmap
  !---------------
  implicit none

  real,                           private :: dkxko
  integer,                        private :: nkxko
  real,                           private ::  kxko

  complex, dimension(:), pointer, private ::  kzko, tkzko

  real   , dimension(:), pointer, private,save :: bin(:),bout(:)
  complex, dimension(:), pointer, save ::  sc(:)

contains
  !----------------------------------------------------------------
  subroutine wemssr_init(max,n)
    real,    intent(in) :: max
    integer, intent(in) :: n
    integer             :: er
    integer             :: i,m
    real                :: kxko

    allocate( kzko(n),stat=er)
    if(er/=0) call seperr('cant allocate  kzko')
    allocate(tkzko(n),stat=er)
    if(er/=0) call seperr('cant allocate tkzko')
    allocate(bin(down%amx%n))
    allocate(bout(down%amx%n))
    allocate(sc(down%amx%n))

    dkxko = max/(n-1)
    nkxko = n
    m=n/max

    !!                    ____________
    !!   kz         1    /     | kx |2          kx
    !! i -- dz = -i -   /  1 - | -- |  dz   0 < -- < 1 
    !!   ko         2 \/       | ko |           ko
    !!
    do i=1,m
       kxko    = 0. +     1.0 * (i  -1)/(  m-1) !! kx/ko=0...1
       kzko(i) = - (0.,+1.) * sqrt(1-kxko**2)
    end do

    !!                    ____________
    !!   kz         1    / | kx |2              kx
    !! i -- dz = -  -   /  | -- | - 1  dz   1 < -- < max 
    !!   ko         2 \/   | ko |               ko
    !!
    do i=m+1,n
       kxko    = 1. + (max-1) * (i-m-1)/(n-m-1) !! kx/ko=1...max
       kzko(i) = - (+1.,0.) * sqrt(kxko**2-1)
    end do
  end subroutine wemssr_init
  !----------------------------------------------------------------
  integer function wem_init() result(st)
    integer :: er
    integer ::i,imax
    real :: hxmax,hymax
#ifdef DEBUG
    character(len=128), parameter :: unit='wem_init'
    if(run%debug) call in(unit)
#endif

    !! init DSR
    call from_param("nkxko",nkxko,2001)
    call from_param( "kxko", kxko,2.)
    call wemssr_init(kxko,nkxko)
     
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function wem_init

  !----------------------------------------------------------------
  integer function wem(iws,izs,ifk,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk
    complex, dimension(:,:,:), pointer :: wfld
    real    :: w, ko
    integer :: ix,iy
    integer :: is,ir,i,iz
    complex :: ikz
    real,dimension(:),pointer :: kx,ky
    real    :: k
#ifdef DEBUG
    character(len=128), parameter :: unit='wem'
    if(run%depar) call in(unit)
#endif

    w  = down%bw%o + (iws-1)*down%bw%d
    ko =  slo%sz(izs,ifk) * w
    kx=>kmap%kx
    ky=>kmap%ky
    if(down%opsign==goB2T) then
       tkzko=conjg(kzko)
    else
       tkzko=kzko
    end if
    ky_loop:do iy=1,down%amy%n
       kx_loop:do ix=1,down%amx%n
          k = sqrt(kx(ix)**2 + ky(iy)**2)
          i =max(1, min(int(1 + k/ko / dkxko) , nkxko))
          ikz= ko*tkzko(i)*var%dzb(izs)
         bin(ix)=real(ikz)
         sc(ix)=cmplx(cos(aimag(ikz)),sin(aimag(ikz)))
!          wfld(ix,iy,iws)= wfld(ix,iy,iws)*exp(ikz)
!          wfld(ix,iy,iws)= wfld(ix,iy,iws)*exp(ikz)
       end do kx_loop
       bout=exp(bin)
       wfld(:,iy,iws)=wfld(:,iy,iws)*bout*sc
    end do ky_loop

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function wem
end module wei_wem
